import sys
import threading
import typing
from concurrent.futures import Future, ThreadPoolExecutor

from applitools.common import VGResource, logger


class ResourceCache(typing.Mapping[typing.Text, VGResource]):
    def __init__(self):
        self.cache_map = {}
        kwargs = {}
        if sys.version_info >= (3, 6):
            kwargs["thread_name_prefix"] = "ResourceCache-Executor"
        self.executor = ThreadPoolExecutor(**kwargs)
        self.lock = threading.RLock()

    def __str__(self):
        return str(self.cache_map)[:25]

    def _process_future(self, url, val):
        if isinstance(val, Future):
            try:
                val = val.result()
            except Exception as e:
                logger.debug(
                    "We got an exception for following URL: {}"
                    "\n  See details below: \n{}".format(url, str(e))
                )
                val = None
            finally:
                self.cache_map[url] = val
        return val

    def __getitem__(self, item):
        with self.lock:
            val = self.cache_map[item]
            self._process_future(item, val)
        return val

    def __setitem__(self, key, value):
        with self.lock:
            self.cache_map[key] = value

    def __len__(self):
        with self.lock:
            return len(self.cache_map)

    def __del__(self):
        with self.lock:
            self.executor.shutdown()

    def __iter__(self):
        return iter(self.cache_map)

    def fetch_and_store(self, url, func_to_run):
        if url in self:
            return self[url]
        if func_to_run:
            self[url] = self.executor.submit(func_to_run, url)
            return True
        return False

    def process_all(self):
        with self.lock:
            return [self[r_url] for r_url in self]



import typing
import uuid

import attr

from applitools.common import TestResults, logger

if typing.TYPE_CHECKING:
    from typing import Callable, Text, Optional


@attr.s(hash=True)
class VGTask(object):
    name = attr.ib()  # type: Text
    func_to_run = attr.ib(hash=False, repr=False)  # type: Callable
    uuid = attr.ib(init=False, repr=False, factory=lambda: str(uuid.uuid4()))

    callback = None
    error_callback = None
    complete_callback = None

    def on_task_succeeded(self, code):
        # type: (Callable) -> VGTask
        self.callback = code
        return self

    def on_task_error(self, code):
        # type: (Callable) -> VGTask
        self.error_callback = code
        return self

    def on_task_completed(self, code):
        # type: (Callable) -> VGTask
        self.complete_callback = code
        return self

    def __call__(self):
        # type: () -> Optional[TestResults]
        logger.debug("%s called %s" % (self.__class__.__name__, self.name))
        res = None
        try:
            if callable(self.func_to_run):
                logger.debug(
                    "VGTask().func_to_run: {}".format(self.func_to_run.__name__)
                )
                res = self.func_to_run()
            if callable(self.callback):
                logger.debug("VGTask().callback: {}".format(self.callback.__name__))
                self.callback(res)
        except Exception as e:
            logger.error("Failed to execute task! \n\t %s" % self.name)
            logger.exception(e)
            if callable(self.error_callback):
                logger.debug(
                    "VGTask().error_callback: {}".format(self.error_callback.__name__)
                )
                self.error_callback(e)
        finally:
            if callable(self.complete_callback):
                logger.debug(
                    "VGTask().complete_callback: {}".format(
                        self.complete_callback.__name__
                    )
                )
                self.complete_callback()
        return res



from abc import abstractmethod

from applitools.common import Point, logger
from applitools.common.utils import ABC
from applitools.core import PositionProvider


class RegionVisibilityStrategy(ABC):
    """
    Encapsulates implementations for providing region visibility during
    check_region.
    """

    @abstractmethod
    def return_to_original_position(self, position_provider):
        # type: (PositionProvider) -> None
        pass

    @abstractmethod
    def move_to_region(self, position_provider, location):
        # type: (PositionProvider, Point) -> None
        pass


class NopRegionVisibilityStrategy(RegionVisibilityStrategy):
    def move_to_region(self, position_provider, location):
        logger.debug("Ignored (no op).")

    def return_to_original_position(self, position_provider):
        logger.debug("Ignored (no op).")


class MoveToRegionVisibilityStrategy(RegionVisibilityStrategy):
    VISIBILITY_OFFSET = 100  # pixels
    _position_memento = None

    def move_to_region(self, position_provider, location):
        # type: (PositionProvider, Point) -> None
        self._position_memento = position_provider.get_state()
        # We set the location to "almost" the location we were asked. This is because
        # sometimes, moving the browser  to the specific pixel where the
        # element begins, causes the element to be slightly out of the viewport.
        location = location - self.VISIBILITY_OFFSET
        location.x = 0 if location.x < 0 else location.x
        location.y = 0 if location.y < 0 else location.y
        position_provider.set_position(location)

    def return_to_original_position(self, position_provider):
        if self._position_memento:
            position_provider.restore_state(self._position_memento)
